import json
import uuid


class RaceHandler:
    """
    Standard race handler.

    You should use this class as a basis for creating your own handler that
    can consume incoming messages, react to race data changes, and send stuff
    back to the race room.
    """
    # This is used by `should_stop` to determine when the handler should quit.
    stop_at = ['cancelled', 'finished']

    def __init__(self, logger, conn):
        """
        Base handler setup.

        Adds the logger, WebSocket connector and prepares some variables.
        """
        self.logger = logger
        self.conn = conn
        self.ws = None
        self.data = {}

    def should_stop(self):
        """
        Determine if the handler should be terminated. This is checked after
        every receieved message.

        By default, checks if the race state matches one of the values in
        `stop_at`.
        """
        return self.data.get('status', {}).get('value') in self.stop_at

    async def begin(self):
        """
        Bot actions to perform when first connecting to a race room.

        Override this method to add an intro for when your bot first appears.
        """
        pass

    async def consume(self, data):
        """
        Standard message consumer. This is called for every message we receive
        from the site.

        This implementation will attempt to find an appropriate method to call
        to handle the incoming data, based on its type. For example if we have
        a "race.data" type message, it will call `self.race_data(data)`.
        """
        msg_type = data.get('type')

        self.logger.info('[%(race)s] Recieved %(msg_type)s' % {
            'race': self.data.get('name'),
            'msg_type': msg_type,
        })

        method = msg_type.replace('.', '_')
        if msg_type and hasattr(self, method):
            await getattr(self, method)(data)
        else:
            self.logger.info(f'No handler for {msg_type}, ignoring.')

    async def end(self):
        """
        Bot actions to perform just before disconnecting from a race room.

        Override this method to add an outro for when your bot leaves.
        """
        pass

    async def error(self, data):
        """
        Consume an incoming "error" type message.

        By default, just raises the message as an exception.
        """
        raise Exception(data.get('errors'))

    async def chat_message(self, data):
        """
        Consume an incoming "chat.message" type message.

        This method assumes a standard bot operation. It checks the first word
        in the message, and if it looks like an exclaimation command like
        "!seed", then it will call the relevant method, i.e.
        `self.ex_seed(args, message)` (where `args` is the remainder of the
        message split up by words, and message is the original message blob).
        """
        message = data.get('message', {})

        if message.get('is_bot') or message.get('is_system'):
            self.logger.info('Ignoring bot/system message.')
            return

        words = message.get('message', '').split(' ')
        if words and words[0][0] == '!':
            method = 'ex_' + words[0][1:]
            args = words[1:]
            if hasattr(self, method):
                self.logger.info('[%(race)s] Calling handler for %(word)s' % {
                    'race': self.data.get('name'),
                    'word': words[0],
                })
                await getattr(self, method)(args, message)

    async def race_data(self, data):
        """
        Consume an incoming "race.data" message.

        By default just updates the `data` attribute on the object. If you
        want to react to race changes, you can override this method to add
        further functionality.
        """
        self.data = data.get('race')

    async def send_message(self, message):
        """
        Send a chat message to the race room.

        `message` should be the message string you want to send.
        """
        await self.ws.send(json.dumps({
            'action': 'message',
            'data': {
                'message': message,
                'guid': str(uuid.uuid4()),
            }
        }))
        self.logger.info('[%(race)s] Message: "%(message)s"' % {
            'race': self.data.get('name'),
            'message': message,
        })

    async def set_raceinfo(self, info, overwrite=False, prefix=True):
        """
        Set the `info` field on the race room's data.

        `info` should be the information you wish to set. By default, this
        method will prefix your information with the existing info, if needed.
        You can change this to suffix with `prefix=False`, or disable this
        behaviour entirely with `overwrite=True`.
        """
        if self.data.get('info') and not overwrite:
            if prefix:
                info = info + ' | ' + self.data.get('info')
            else:
                info = self.data.get('info') + ' | ' + info

        await self.ws.send(json.dumps({
            'action': 'setinfo',
            'data': {'info': info}
        }))
        self.logger.info('[%(race)s] Set info: "%(info)s"' % {
            'race': self.data.get('name'),
            'info': info,
        })

    async def handle(self):
        """
        Low-level handler for the race room. This will loop over the websocket,
        processing any messages that come in.
        """
        self.logger.info('[%(race)s] Handler started' % {
            'race': self.data.get('name'),
        })
        async with self.conn as ws:
            self.ws = ws
            await self.begin()
            async for message in self.ws:
                data = json.loads(message)
                await self.consume(data)
                if self.should_stop():
                    await self.end()
                    break



import asyncio
import json
from datetime import timedelta
from functools import partial

import requests
import websockets

from .handler import RaceHandler


class Bot:
    """
    The racetime.gg bot class.

    This bot uses event loops to connect to multiple race rooms at once. Each
    room is assigned a handler object, which you can determine with the
    `get_handler_class`/`get_handler_kwargs` methods.

    When implementing your own bot, you will need to specify what handler you
    wish to use, as the default `RaceHandler` will not actually do anything
    other than connect to the room and log the messages.
    """
    racetime_host = 'racetime.gg'
    racetime_secure = True
    scan_races_every = timedelta(seconds=30)
    gather_tasks_every = timedelta(seconds=30)

    def __init__(self, category_slug, client_id, client_secret, logger,
                 ssl_context=None):
        """
        Bot constructor.
        """
        self.logger = logger
        self.category_slug = category_slug
        self.ssl_context = ssl_context

        self.loop = asyncio.get_event_loop()
        self.last_scan = None
        self.races = {}
        self.handlers = {}

        self.access_token = self.authorize(client_id, client_secret)

    def get_handler_class(self):
        """
        Returns the handler class for races. Each race the bot finds will have
        its own handler object of this class.
        """
        return RaceHandler

    def get_handler_kwargs(self, ws_conn):
        """
        Returns a dict of keyword arguments to be passed into the handler class
        when instantiated.

        Override this method if you need to pass additional kwargs to your race
        handler.
        """
        return {
            'logger': self.logger,
            'conn': ws_conn,
        }

    def should_handle(self, race_data):
        """
        Determine if a race we've found should be handled by this bot or not.

        Returns True if the race should have a handler created for it, False
        otherwise.
        """
        return race_data.get('status', {}).get('value') not in RaceHandler.stop_at

    def authorize(self, client_id, client_secret):
        """
        Get an OAuth2 token from the authentication server.
        """
        resp = requests.post(self.http_uri('/o/token'), {
            'client_id': client_id,
            'client_secret': client_secret,
            'grant_type': 'client_credentials',
        })
        data = json.loads(resp.content)
        token = data.get('access_token')
        if not token:
            raise Exception('Unable to retrieve access token.')
        return token

    def create_handler(self, race_data):
        """
        Create a new WebSocket connection and set up a handler object to manage
        it.
        """
        ws_conn = websockets.connect(
            self.ws_uri(race_data.get('websocket_bot_url')),
            extra_headers={
                'Authorization': 'Bearer ' + self.access_token,
            },
            ssl=self.ssl_context,
        )

        cls = self.get_handler_class()
        kwargs = self.get_handler_kwargs(ws_conn)

        handler = cls(**kwargs)

        self.logger.info(
            'Created handler for %(race)s'
            % {'race': race_data.get('name')}
        )

        return handler

    async def gather_tasks(self):
        """
        Examines the current races, and creates a handler and task for any race
        that should be handled but currently isn't.

        This method runs in a constant loop, checking for new races every few
        seconds.
        """
        tasks = {}

        def done(task_name, *args):
            del tasks[task_name]

        while True:
            for name, summary_data in self.races.items():
                if name not in tasks:
                    resp = requests.get(self.http_uri(summary_data.get('data_url')))
                    race_data = json.loads(resp.content)
                    if self.should_handle(race_data):
                        handler = self.create_handler(race_data)
                        tasks[name] = self.loop.create_task(handler.handle())
                        tasks[name].add_done_callback(partial(done, name))
                        self.logger.debug(tasks[name])
                    else:
                        self.logger.info(
                            'Ignoring %(race)s by configuration.'
                            % {'race': race_data.get('name')}
                        )
            await asyncio.sleep(self.gather_tasks_every.seconds)

    async def refresh_races(self):
        """
        Retrieve current race information from the category detail API
        endpoint, and populate the race list.

        This method runs in a constant loop, checking for new races every few
        seconds.
        """
        while True:
            self.logger.info('Refresh races')
            resp = requests.get(self.http_uri(f'/{self.category_slug}/data'))
            data = json.loads(resp.content)
            self.races = {}
            for race in data.get('current_races', []):
                self.races[race.get('name')] = race
            await asyncio.sleep(self.scan_races_every.seconds)

    def run(self):
        """
        Run the bot. Creates an event loop then iterates over it forever.
        """
        self.loop.create_task(self.refresh_races())
        self.loop.create_task(self.gather_tasks())
        self.logger.info('Running event loop')
        self.loop.run_forever()

    def http_uri(self, url):
        """
        Generate a HTTP/HTTPS URI from the given URL path fragment.
        """
        return self.uri(
            proto='https' if self.racetime_secure else 'http',
            url=url,
        )

    def ws_uri(self, url):
        """
        Generate a WS/WSS URI from the given URL path fragment.
        """
        return self.uri(
            proto='wss' if self.racetime_secure else 'ws',
            url=url,
        )

    def uri(self, proto, url):
        """
        Generate a URI from the given protocol and URL path fragment.
        """
        return '%(proto)s://%(host)s%(url)s' % {
            'proto': proto,
            'host': self.racetime_host,
            'url': url,
        }



from setuptools import find_packages, setup


setup(
    name='racetime-bot',
    description='Foundation system for creating chat bots for racetime.gg',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    url='https://github.com/deains/racetime-bot',
    project_urls={
        'Source': 'https://github.com/deains/racetime-bot',
    },
    version='1.0.1',
    install_requires=[
        'asgiref',
        'requests',
        'websockets',
    ],
    packages=find_packages(),
)
